import torch
import json

from .. import __version__, STFTParams
from ..separation.base import SeparationException
from ..datasets import transforms as tfm
from ..evaluation import BSSEvalV4, BSSEvalScale


class SafeModelLoader(object):
    """
    Loads a nussl model and populates the metadata with defaults if
    """

    # Expected topology of the model's metadata as of nussl version 1.1.3
    _v1_1_3_metadata = {
        'config': {
            'connections': list,
            'modules': dict,
            'name': str,
            'output': list
        },
        'evaluation': None,
        'loss_dictionary': dict,
        'num_channels': int,
        'nussl_version': str,
        'optimizer': {
            'name': str,
            'params': dict
        },
        'sample_rate': int,
        'stft_params': STFTParams,
        'train_dataset': {
            'folder': str,
            'name': str,
            'num_channels': int,
            'sample_rate': int,
            'stft_params': STFTParams,
            'transforms': tfm.Compose
        },
        'trainer.state.epoch_history': dict,
        'trainer.state_dict': {
            'epoch': int,
            'epoch_length': int,
            'max_epochs': int,
            'metrics': dict,
            'output': dict,
            'seed': None
        },
        'val_dataset': {
            'folder': str,
            'name': str,
            'num_channels': int,
            'sample_rate': int,
            'stft_params': STFTParams,
            'transforms': tfm.Compose
        },
    }

    expected_metadata = _v1_1_3_metadata

    def __init__(self):
        """
        Drop in replacement for torch.load(). Will load a model and return
        a model_dict that is populated
        Args:
            model_path (str): Path to a nussl-saved model.
            device (str):
            expected_eval (str): Either 'BSSEvalScale' or 'BSSEvalV4'. Will
                look for & populate missing eval keys with the format of either
                of those eval methods.

        Returns:
            model_dict (dict):

        """
        self.current_version = __version__
        self.eval = None

    def load(self, model_path, device='cpu', expected_eval='BssEvalScale'):
        model_dict = torch.load(model_path, map_location=device)
        metadata = model_dict['metadata']
        metadata = self._get_moved(metadata, model_dict, model_path)
        self.eval = expected_eval
        metadata = self._validate_and_populate(metadata)
        metadata['config'] = json.dumps(metadata['config'])
        model_dict['metadata'] = metadata
        return model_dict

    @staticmethod
    def _get_moved(metadata, model_dict, model_path):
        model_dict_version = model_dict.get('nussl_version', None)
        metadata_version = metadata.get('nussl_version', None)

        if metadata_version is not None:
            saved_version = metadata_version
        else:
            saved_version = model_dict_version

        if saved_version is None:
            raise SeparationException(f"Failed loading model. Expected to find "
                                      f"'nussl_version' in {model_path}.")

        metadata['nussl_version'] = saved_version

        if 'config' in model_dict:
            metadata['config'] = json.loads(model_dict['config'])

        if 'transforms' in metadata:
            if 'train_dataset' not in metadata:
                metadata['train_dataset'] = {}
            metadata['train_dataset']['transforms'] = metadata['transforms']

        return metadata

    def _load_eval(self, eval_dict):
        """Helper function to load eval dictionary safely."""
        if self.eval.lower() == 'bssevalv4':
            keys = BSSEvalV4.keys
        else:
            keys = BSSEvalScale.keys

        stats_keys = ['mean', 'median', 'std']

        result = {}
        for k in keys:
            if k not in eval_dict:
                stats = {s: 'UNAVAILABLE' for s in stats_keys}
            else:
                stats = {}
                for s in stats_keys:
                    if s in eval_dict[k]:
                        stats[s] = eval_dict[k][s]
                    else:
                        stats[s] = 'UNAVAILABLE'
            result[k] = stats
        return result

    @staticmethod
    def _load_types(expected_type, key, val):
        """Safe load for values where the value is a type in self.expected_metadata"""
        if val is not None:
            if type(val) != expected_type:
                raise SeparationException(f'Expected type {expected_type} '
                                          f'for key {key} but got {type(val)}!')
            return val
        else:
            return 'UNAVAILABLE'

    def _validate_and_populate(self, received):
        """Safe load for metadata according to the expected metadata."""
        result = {}
        for key, expected_val in self.expected_metadata.items():
            val = received.get(key, None)

            if key == 'evaluation':
                eval_dict = received.get('evaluation', {})
                result['evaluation'] = self._load_eval(eval_dict)
            elif type(expected_val) == type:
                result[key] = self._load_types(expected_val, key, val)

            elif type(expected_val) == dict:
                next_dict = {} if val is None else val
                sub_result = {}
                for sub_key, type_ in expected_val.items():
                    sub_val = next_dict.get(sub_key, None)
                    sub_result[sub_key] = self._load_types(type_, sub_key, sub_val)
                result[key] = sub_result

        return result


import numpy as np
import librosa

from ..base import ClusteringSeparationBase, NMFMixin


class TimbreClustering(ClusteringSeparationBase, NMFMixin):
    """
    Implements separation by timbre via NMF with MFCC clustering. The
    steps are:

    1. Factorize the magnitude spectrogram of the mixture with NMF.
    2. Take MFCC coefficients of each component.
    3. Express each time-frequency bin as a combination of components.
    4. The features for each time-frequency bin are the weighted combination
       of the MFCCs of each component.
    5. Cluster each time-frequency bin based on these features.
    
    Args:
        input_audio_signal (AudioSignal): Signal to separate.
        n_components (int): Number of components to use in the NMF
          model. Corresponds to number of spectral templates.
        n_mfcc (int): Number of MFCC coefficients to use. Defaults to 13.
        nmf_args (dict): Dictionary containing keyword arguments for `NMFMixin.fit`.
        kwargs (dict): Extra keyword arguments are passed to ClusteringSeparationBase.
    """

    def __init__(self, input_audio_signal, num_sources, n_components, n_mfcc=13,
                 nmf_kwargs=None, **kwargs):
        self.n_components = n_components
        self.nmf_kwargs = {} if nmf_kwargs is None else nmf_kwargs
        self.n_mfcc = n_mfcc

        super().__init__(
            input_audio_signal=input_audio_signal,
            num_sources=num_sources, **kwargs)

    def extract_features(self):
        model, components, activations = self.fit(
            self.audio_signal, self.n_components, **self.nmf_kwargs)
        mel = librosa.feature.melspectrogram(
            S=(components.T ** 2), sr=self.sample_rate)
        mfcc = librosa.feature.mfcc(S=mel, n_mfcc=self.n_mfcc).T

        activations = activations.reshape(self.n_components, -1, 1)
        components = components.reshape(self.n_components, 1, -1)

        expansion = activations @ components
        expansion = expansion.transpose()
        features = expansion @ mfcc

        features = features.reshape(self.stft.shape + (-1,))
        norm = np.linalg.norm(features, axis=-1, keepdims=True)
        features = features / (norm + 1e-8)

        return features



"""
Deep clustering
---------------

.. autoclass:: nussl.separation.deep.DeepClustering
    :autosummary:

Deep mask estimation
--------------------

.. autoclass:: nussl.separation.deep.DeepMaskEstimation
    :autosummary:

Deep audio estimation
---------------------

.. autoclass:: nussl.separation.deep.DeepAudioEstimation
    :autosummary:

"""

from .deep_clustering import DeepClustering
from .deep_mask_estimation import DeepMaskEstimation
from .deep_audio_estimation import DeepAudioEstimation



from ...unfold import GaussianMixtureTorch
from .filter_bank import FilterBank, STFT, LearnedFilterBank
from .blocks import (
    AmplitudeToDB,
    Alias,
    ShiftAndScale, 
    BatchNorm,
    InstanceNorm,
    GroupNorm,
    LayerNorm,
    MelProjection,
    Embedding,
    Mask,
    Split,
    Expand,
    Concatenate,
    RecurrentStack,
    ConvolutionalStack2D,
    DualPathBlock,
    DualPath,
)



import copy

import numpy as np
import sklearn

from .. import SeparationBase
from ... import AudioSignal
from ...core import utils


class ICA(SeparationBase):
    """
    Separate sources using the Independent Component Analysis, given 
    observations of the audio scene. nussl's ICA is a wrapper for sci-kit learn's 
    implementation of FastICA, and provides a way to interop between 
    nussl's :ref:`AudioSignal` objects and FastICA.

    References:
        `sci-kit learn FastICA <http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.fastica.html>`_

    Args:
        audio_signals: list of AudioSignal objects containing the observations 
          of the mixture. Will be converted into a single multichannel AudioSignal.
        
        max_iterations (int): Max number of iterations to run ICA for. Defaults to 200.

        **kwargs: Additional keyword arguments that will be passed to 
          `sklearn.decomposition.FastICA`
    """

    def __init__(self, audio_signals, max_iterations=200, **kwargs):
        super().__init__(input_audio_signal=audio_signals)

        # FastICA setup attributes
        self.num_components = self.audio_signal.num_channels
        self.kwargs = kwargs
        self.max_iterations = max_iterations

        # Results attributes
        self.estimated_sources = None
        self.estimated_mixing_params = None
        self.mean = None
        self.ica_output = None

    @property
    def audio_signal(self):
        """
        Copy of AudioSignal that is made on initialization.
        """
        return self._audio_signal

    @audio_signal.setter
    def audio_signal(self, audio_signals):
        """
        Takes a list of audio signals and constructs a single multichannel audio signal
        object.

        Args:
            audio_signal (list or AudioSignal): Either a multichannel audio signal,
              or a list of AudioSignals containing the observations.
        """

        if isinstance(audio_signals, list):
            audio_signals = utils.verify_audio_signal_list_strict(audio_signals)
            audio_data = np.vstack([s.audio_data for s in audio_signals])
            audio_signal = audio_signals[0].make_copy_with_audio_data(audio_data)
            self._audio_signal = audio_signal
        elif isinstance(audio_signals, AudioSignal):
            self._audio_signal = copy.deepcopy(audio_signals)

    def run(self):
        ica = sklearn.decomposition.FastICA(
            n_components=self.num_components, max_iter=self.max_iterations,
            **self.kwargs)

        # save for normalizing the estimated signals
        max_input_amplitude = np.max(np.abs(self.audio_signal.audio_data))

        # run ICA
        ica_output = ica.fit_transform(self.audio_signal.audio_data.T).T

        # now normalize the estimated signals
        max_output_amplitude = np.max(np.abs(ica_output))
        ica_output /= max_output_amplitude
        ica_output *= max_input_amplitude

        # store the resultant computations
        self.estimated_mixing_params = ica.mixing_
        self.mean = ica.mean_
        self.ica_output = ica_output

        return self.ica_output

    def make_audio_signals(self):
        estimated_sources = [
            AudioSignal(
                audio_data_array=self.ica_output[i, :],
                sample_rate=self.audio_signal.sample_rate)
            for i in range(self.ica_output.shape[0])
        ]
        return estimated_sources



import copy

import numpy as np
import torch

from .. import SeparationBase, SeparationException
from ... import AudioSignal


class Projet(SeparationBase):
    """
    Implements the PROJET algorithm for spatial audio separation using projections.
    This implementation uses PyTorch to speed up computation considerably. PROJET
    does the following steps:

    1. Project the complex stereo STFT onto multiple angles and delay via
       projection and delay matrix transformations.
    2. Initialize the parameters of the system to "remix" these projections along with
       PSDs of the sources such that they try to reconstruct the original stereo mixture.
    3. Find the optimal parameters via multiplicative update rules for P and for Q.
    4. Use the discovered parameters to isolate the sources via spatial cues.

    This implementation considers BOTH panning and delays when isolating sources.
    PROJET is not a masking based method, it estimates the sources directly by
    projecting the complex STFT.

    Args:
          input_audio_signal (AudioSignal): Audio signal to separate.
          num_sources (int): Number of source to separate.
          estimates (list of AudioSignal):  initial estimates for the separated sources
            if available. These will be used to initialize the update algorithm. So
            one could (for example), run FT2D on a signal and then refine the estimates
            using PROJET. Defaults to None (randomly initialize P).
          num_iterations (int, optional): Number of iterations to do for the update 
            rules for P and Q. Defaults to 50.
          maximum_delay_in_samples (int, optional): Maximum delay in samples that you are
            willing to consider in the projection matrices. Defaults to 20.
          location_set_panning (int, optional): How many locations in panning you are 
            willing to consider. Defaults to 30.
          location_set_delay (int, optional): How many delays you are willing to 
            consider. Defaults to 17.
          projection_set_panning (int, optional): How many projections you are willing
            use in panning-space. Defaults to 10.
          projection_set_delay (int, optional): How many delays you are willing to project
            the mixutre onto in panning-space. Defaults to 9.
          beta (int, optional):  Beta in beta divergence. See Table 1 in [1]. Defaults to 1.
          alpha (int, optional): Power to raise each power spectral density estimate of each
            source to. Defaults to 1.
          device (str, optional): Device to use when performing update rules. 'cuda' will
            be fastest, if available. Defaults to 'cpu'.

    References:

    [1] Fitzgerald, Derry, Antoine Liutkus, and Roland Badeau. 
        "Projection-based demixing of spatial audio." 
        IEEE/ACM Transactions on Audio, Speech, and Language 
        Processing 24.9 (2016): 1560-1572.

    [2] Fitzgerald, Derry, Antoine Liutkus, and Roland Badeau. 
        "Projet—spatial audio separation using projections." 2016 IEEE International 
        Conference on Acoustics, Speech and Signal Processing (ICASSP). IEEE, 2016.
    """

    def __init__(self, input_audio_signal, num_sources, estimates=None, num_iterations=50,
                 maximum_delay_in_samples=20, location_set_panning=30, location_set_delay=17,
                 projection_set_panning=10, projection_set_delay=9, beta=1, alpha=1, device='cpu'):

        self.num_sources = num_sources
        self.alpha = alpha
        self.beta = beta
        self.num_iterations = num_iterations
        self.eps = 1e-8
        self.device = device
        self.projection_set_panning = projection_set_panning
        self.projection_set_delay = projection_set_delay
        self.location_set_panning = location_set_panning
        self.location_set_delay = location_set_delay
        self.maximum_delay_in_samples = maximum_delay_in_samples
        self.projection_set = None
        self.inverse_projection_set = None
        self.reconstructions = None

        super().__init__(input_audio_signal=input_audio_signal)

        if estimates is None:
            self.P = None
        else:
            if len(estimates) != self.num_sources:
                raise SeparationException(
                    "Number of estimates must be equal to num_sources!")
            psds = []
            for e in estimates:
                _e = copy.deepcopy(e)
                _e.to_mono()
                _e.stft_params = self.audio_signal.stft_params
                psds.append(np.abs(_e.stft()))

            self.P = np.stack(psds, axis=-1)

    def _preprocess_audio_signal(self):
        self.stft = self.audio_signal.stft()
        F, T, I = self.stft.shape

        self.device = 'cpu' if not torch.cuda.is_available() else self.device

        pannings = np.linspace(0, np.pi / 2, self.location_set_panning)
        delays = np.linspace(
            -self.maximum_delay_in_samples,
            self.maximum_delay_in_samples,
            self.location_set_delay
        )
        self.location_set = self.create_panning_delay_set(pannings, delays, F, I)

        pannings = np.linspace(-np.pi / 2, 0, self.projection_set_panning)
        delays = np.linspace(
            -self.maximum_delay_in_samples,
            self.maximum_delay_in_samples,
            self.projection_set_delay
        )
        self.projection_set = self.create_panning_delay_set(pannings, delays, F, I)
        self.inverse_projection_set = np.linalg.pinv(self.projection_set)

        self.V, self.complex_projections = self.create_projections()
        self.K = self.create_k_matrix()

    @staticmethod
    def create_panning_delay_set(pannings, delays, F, I):
        panning_delay_set = np.zeros(
            (F, len(pannings), len(delays), I), dtype='complex'
        )

        for i, delay in enumerate(delays):
            phase_change = np.exp(
                -1j * 2 * np.pi * np.linspace(0, 0.5, F) * delay
            )
            panning_delay_set[:, :, i, 0] = np.outer(
                np.ones(F), np.cos(pannings)
            )
            panning_delay_set[:, :, i, 1] = np.outer(
                phase_change, np.sin(pannings)
            )
        return panning_delay_set

    def create_projections(self):
        F = self.stft.shape[0]
        inner_dim = self.projection_set.shape[-1]
        V = (
                self.projection_set.reshape(F, -1, inner_dim) @
                self.stft.reshape(F, -1, inner_dim).transpose(0, 2, 1)
        )
        V = V.reshape((F,) + self.projection_set.shape[1:-1] + (-1,))
        return self._convert_to_tensor(np.abs(V) ** self.alpha), V

    def _convert_to_tensor(self, data):
        tensor = torch.from_numpy(data)
        return tensor.to(self.device)

    @staticmethod
    def _convert_to_numpy(data):
        array = data.cpu().data.numpy()
        return array

    def create_k_matrix(self):
        F = self.stft.shape[0]
        inner_dim = self.location_set.shape[-1]

        K = (
                self.projection_set.reshape(F, -1, inner_dim) @
                self.location_set.reshape(F, -1, inner_dim).transpose(0, 2, 1)
        )

        K = np.abs(K.reshape(
            (F,) + self.projection_set.shape[1:-1] + self.location_set.shape[1:-1])) ** self.alpha
        return self._convert_to_tensor(K)

    def initialize_parameters(self):
        F, T, I = self.stft.shape
        P = np.abs(np.random.randn(F, T, self.num_sources))
        Q = np.abs(np.random.randn(*self.location_set.shape[1:3], self.num_sources))
        return self._convert_to_tensor(P), self._convert_to_tensor(Q)

    def _update_sigma(self, P, Q, KQ):
        F = self.stft.shape[0]
        inner_dim = KQ.shape[-1]
        sigma = (
                KQ.reshape(F, -1, inner_dim) @
                P.transpose(2, 1)
        )

        sigma = sigma.reshape(
            P.shape[0], KQ.shape[1], KQ.shape[2], P.shape[1])
        return sigma

    def _update_P(self, P, sigma, KQ):
        F = self.stft.shape[0]
        temps = [
            (sigma ** (self.beta - 2)) * self.V,
            sigma ** (self.beta - 1)
        ]
        inner_dim = KQ.shape[1] * KQ.shape[2]
        P_num_denom = [
            self.eps + (
                    KQ.reshape(F, inner_dim, -1).transpose(2, 1) @
                    temp.reshape(F, inner_dim, -1)
            )
            for temp in temps
        ]
        P_update = (P_num_denom[0] / P_num_denom[1]).transpose(2, 1)
        return P * P_update

    def _update_Q(self, P, sigma, Q):
        F = self.stft.shape[0]
        temps = [
            (sigma ** (self.beta - 2)) * self.V,
            sigma ** (self.beta - 1)
        ]
        inner_dim = self.K.shape[1] * self.K.shape[2]
        Q_num_denom = [
            self.K.reshape(F, inner_dim, -1).transpose(2, 1) @
            (temp.reshape(F, -1, P.shape[1]) @ P)
            for temp in temps
        ]
        Q_num_denom = [
            x.reshape(F, *Q.shape).sum(dim=0) for x in Q_num_denom
        ]
        Q_update = Q_num_denom[0] / Q_num_denom[1]
        return Q * Q_update

    def _get_kq(self, Q):
        F = self.stft.shape[0]
        # get KQ
        inner_dim = Q.shape[0] * Q.shape[1]
        KQ = (
                self.K.reshape(-1, inner_dim) @ Q.reshape(inner_dim, -1)
        )
        KQ = KQ.reshape(
            F, self.K.shape[1], self.K.shape[2], Q.shape[-1])
        return KQ

    def _update(self, P, Q):
        KQ = self._get_kq(Q)
        sigma = self._update_sigma(P, Q, KQ)
        P = self._update_P(P, sigma, KQ)
        sigma = self._update_sigma(P, Q, KQ)
        Q = self._update_Q(P, sigma, Q)
        return P, Q

    def run(self):
        P, Q = self.initialize_parameters()
        for i in range(self.num_iterations):
            P, Q = self._update(P, Q)

        KQ = self._get_kq(Q)
        KQ = KQ.reshape(KQ.shape[0], -1, 1, KQ.shape[-1])
        sigma_j = KQ / P[:, None, ...]
        sigma_j = sigma_j / (self.eps + sigma_j.sum(dim=-1)[..., None])
        sigma_j = self._convert_to_numpy(sigma_j)

        self.projection_set = self.projection_set.reshape(
            self.projection_set.shape[0],
            self.projection_set.shape[1] * self.projection_set.shape[2],
            self.projection_set.shape[-1]
        )

        self.inverse_projection_set = np.linalg.pinv(self.projection_set)

        cf_j = (
                (self.projection_set @
                 self.stft.transpose(0, 2, 1))[..., None]
                * sigma_j
        )
        shape = cf_j.shape

        reconstructions = (
                self.inverse_projection_set @
                cf_j.reshape(
                    cf_j.shape[0],
                    cf_j.shape[1],
                    -1
                )
        )
        reconstructions = reconstructions.reshape(
            shape[0], self.stft.shape[-1], -1, shape[-1]
        )

        self.reconstructions = np.swapaxes(reconstructions, 1, 2)

        return reconstructions

    def make_audio_signals(self):
        estimates = []
        for j in range(self.reconstructions.shape[-1]):
            estimate_stft = self.reconstructions[..., j]
            estimate = self.audio_signal.make_copy_with_stft_data(estimate_stft)
            estimate.istft()
            estimates.append(estimate)
        return estimates



"""
Functions that make it easy to build commonly used source separation architectures.
Currently contains mask inference, deep clustering, and chimera networks that are
based on recurrent neural networks. These functions are a good place to start when
creating your own network toplogies. Since there can be dependencies between layers
depending on input size, it's good to work this out in a function like those below.
"""


def build_recurrent_mask_inference(num_features, hidden_size, num_layers, bidirectional,
                                   dropout, num_sources, mask_activation, num_audio_channels=1,
                                   rnn_type='lstm', normalization_class='BatchNorm',
                                   normalization_args=None, mix_key='mix_magnitude'):
    """
    Builds a config for a mask inference network that can be passed to 
    SeparationModel. This mask inference network uses a recurrent neural network (RNN)
    to process the input representation.
    
    Args:
        num_features (int): Number of features in the input spectrogram (usually means
          window length of STFT // 2 + 1.)
        hidden_size (int): Hidden size of the RNN.
        num_layers (int): Number of layers in the RNN.
        bidirectional (int): Whether the RNN is bidirectional.
        dropout (float): Amount of dropout to be used between layers of RNN.
        num_sources (int): Number of sources to create masks for. 
        mask_activation (list of str): Activation of the mask ('sigmoid', 'softmax', etc.). 
          See ``nussl.ml.networks.modules.Embedding``. 
        num_audio_channels (int): Number of audio channels in input (e.g. mono or stereo).
          Defaults to 1.
        rnn_type (str, optional): RNN type, either 'lstm' or 'gru'. Defaults to 'lstm'.
        normalization_class (str, optional): Type of normalization to apply, either
          'InstanceNorm' or 'BatchNorm'. Defaults to 'BatchNorm'.
        normalization_args (dict, optional): Args to normalization class, optional.
        mix_key (str, optional): The key to look for in the input dictionary that contains
          the mixture spectrogram. Defaults to 'mix_magnitude'.
    
    Returns:
        dict: A recurrent mask inference network configuration that can be passed to
          SeparationModel.
    """
    normalization_args = {} if normalization_args is None else normalization_args

    # define the building blocks
    modules = {
        mix_key: {},
        'log_spectrogram': {
            'class': 'AmplitudeToDB',
        },
        'normalization': {
            'class': normalization_class,
            'args': normalization_args,
        },
        'recurrent_stack': {
            'class': 'RecurrentStack',
            'args': {
                'num_features': num_features,
                'hidden_size': hidden_size,
                'num_layers': num_layers,
                'bidirectional': bidirectional,
                'dropout': dropout,
                'rnn_type': rnn_type
            }
        },
        'mask': {
            'class': 'Embedding',
            'args': {
                'num_features': num_features,
                'hidden_size': hidden_size * 2 if bidirectional else hidden_size,
                'embedding_size': num_sources,
                'activation': mask_activation,
                'num_audio_channels': num_audio_channels
            }
        },
        'estimates': {
            'class': 'Mask',
        },
    }

    # define the topology
    connections = [
        ['log_spectrogram', [mix_key, ]],
        ['normalization', ['log_spectrogram', ]],
        ['recurrent_stack', ['normalization', ]],
        ['mask', ['recurrent_stack', ]],
        ['estimates', ['mask', mix_key]]
    ]

    # define the outputs
    output = ['estimates', 'mask']

    # put it together
    config = {
        'name': 'MaskInference',
        'modules': modules,
        'connections': connections,
        'output': output
    }

    return config


def build_recurrent_dpcl(num_features, hidden_size, num_layers, bidirectional,
                         dropout, embedding_size, embedding_activation, num_audio_channels=1,
                         rnn_type='lstm', normalization_class='BatchNorm', 
                         normalization_args=None, mix_key='mix_magnitude'):
    """
    Builds a config for a deep clustering network that can be passed to 
    SeparationModel. This deep clustering network uses a recurrent neural network (RNN)
    to process the input representation.
    
    Args:
        num_features (int): Number of features in the input spectrogram (usually means
          window length of STFT // 2 + 1.)
        hidden_size (int): Hidden size of the RNN.
        num_layers (int): Number of layers in the RNN.
        bidirectional (int): Whether the RNN is bidirectional.
        dropout (float): Amount of dropout to be used between layers of RNN.
        embedding_size (int): Embedding dimensionality of the deep clustering network. 
        embedding_activation (list of str): Activation of the embedding ('sigmoid', 'softmax', etc.). 
          See ``nussl.ml.networks.modules.Embedding``. 
        num_audio_channels (int): Number of audio channels in input (e.g. mono or stereo).
          Defaults to 1.
        rnn_type (str, optional): RNN type, either 'lstm' or 'gru'. Defaults to 'lstm'.
        normalization_class (str, optional): Type of normalization to apply, either
          'InstanceNorm' or 'BatchNorm'. Defaults to 'BatchNorm'.
        normalization_args (dict, optional): Args to normalization class, optional.
        mix_key (str, optional): The key to look for in the input dictionary that contains
          the mixture spectrogram. Defaults to 'mix_magnitude'.
    
    Returns:
        dict: A recurrent deep clustering network configuration that can be passed to
          SeparationModel.
    """
    normalization_args = {} if normalization_args is None else normalization_args

    # define the building blocks
    modules = {
        mix_key: {},
        'log_spectrogram': {
            'class': 'AmplitudeToDB',
        },
        'normalization': {
            'class': normalization_class,
            'args': normalization_args,
        },
        'recurrent_stack': {
            'class': 'RecurrentStack',
            'args': {
                'num_features': num_features,
                'hidden_size': hidden_size,
                'num_layers': num_layers,
                'bidirectional': bidirectional,
                'dropout': dropout,
                'rnn_type': rnn_type
            }
        },
        'embedding': {
            'class': 'Embedding',
            'args': {
                'num_features': num_features,
                'hidden_size': hidden_size * 2 if bidirectional else hidden_size,
                'embedding_size': embedding_size,
                'activation': embedding_activation,
                'num_audio_channels': num_audio_channels
            }
        },
    }

    # define the topology
    connections = [
        ['log_spectrogram', ['mix_magnitude', ]],
        ['normalization', ['log_spectrogram', ]],
        ['recurrent_stack', ['normalization', ]],
        ['embedding', ['recurrent_stack', ]],
    ]

    # define the outputs
    output = ['embedding']

    # put it together
    config = {
        'name': 'DeepClustering',
        'modules': modules,
        'connections': connections,
        'output': output
    }

    return config


def build_recurrent_chimera(num_features, hidden_size, num_layers, bidirectional,
                            dropout, embedding_size, embedding_activation, num_sources,
                            mask_activation, num_audio_channels=1, 
                            rnn_type='lstm', normalization_class='BatchNorm',
                            normalization_args=None, mix_key='mix_magnitude'):
    """
    Builds a config for a Chimera network that can be passed to SeparationModel. 
    Chimera networks are so-called because they have two "heads" which can be trained
    via different loss functions. In traditional Chimera, one head is trained using a
    deep clustering loss while the other is trained with a mask inference loss. 
    This Chimera network uses a recurrent neural network (RNN) to process the input 
    representation.
    
    Args:
        num_features (int): Number of features in the input spectrogram (usually means
          window length of STFT // 2 + 1.)
        hidden_size (int): Hidden size of the RNN.
        num_layers (int): Number of layers in the RNN.
        bidirectional (int): Whether the RNN is bidirectional.
        dropout (float): Amount of dropout to be used between layers of RNN.
        embedding_size (int): Embedding dimensionality of the deep clustering network. 
        embedding_activation (list of str): Activation of the embedding ('sigmoid', 'softmax', etc.). 
          See ``nussl.ml.networks.modules.Embedding``. 
        num_sources (int): Number of sources to create masks for. 
        mask_activation (list of str): Activation of the mask ('sigmoid', 'softmax', etc.). 
          See ``nussl.ml.networks.modules.Embedding``. 
        num_audio_channels (int): Number of audio channels in input (e.g. mono or stereo).
          Defaults to 1.
        rnn_type (str, optional): RNN type, either 'lstm' or 'gru'. Defaults to 'lstm'.
        normalization_class (str, optional): Type of normalization to apply, either
          'InstanceNorm' or 'BatchNorm'. Defaults to 'BatchNorm'.
        normalization_args (dict, optional): Args to normalization class, optional.
        mix_key (str, optional): The key to look for in the input dictionary that contains
          the mixture spectrogram. Defaults to 'mix_magnitude'.
    
    Returns:
        dict: A recurrent Chimera network configuration that can be passed to
          SeparationModel.
    """
    normalization_args = {} if normalization_args is None else normalization_args

    # define the building blocks
    modules = {
        mix_key: {},
        'log_spectrogram': {
            'class': 'AmplitudeToDB',
        },
        'normalization': {
            'class': normalization_class,
            'args': normalization_args,
        },
        'recurrent_stack': {
            'class': 'RecurrentStack',
            'args': {
                'num_features': num_features,
                'hidden_size': hidden_size,
                'num_layers': num_layers,
                'bidirectional': bidirectional,
                'dropout': dropout,
                'rnn_type': rnn_type
            }
        },
        'embedding': {
            'class': 'Embedding',
            'args': {
                'num_features': num_features,
                'hidden_size': hidden_size * 2 if bidirectional else hidden_size,
                'embedding_size': embedding_size,
                'activation': embedding_activation,
                'num_audio_channels': num_audio_channels
            }
        },
        'mask': {
            'class': 'Embedding',
            'args': {
                'num_features': num_features,
                'hidden_size': hidden_size * 2 if bidirectional else hidden_size,
                'embedding_size': num_sources,
                'activation': mask_activation,
                'num_audio_channels': num_audio_channels
            }
        },
        'estimates': {
            'class': 'Mask',
        },
    }

    # define the topology
    connections = [
        ['log_spectrogram', ['mix_magnitude', ]],
        ['normalization', ['log_spectrogram', ]],
        ['recurrent_stack', ['normalization', ]],
        ['embedding', ['recurrent_stack', ]],
        ['mask', ['recurrent_stack', ]],
        ['estimates', ['mask', 'mix_magnitude']]
    ]

    # define the outputs
    output = ['embedding', 'estimates', 'mask']

    # put it together
    config = {
        'name': 'Chimera',
        'modules': modules,
        'connections': connections,
        'output': output
    }

    return config

def build_open_unmix_like(num_features, hidden_size, num_layers, 
                          bidirectional, dropout, num_sources,
                          num_audio_channels=1, add_embedding=False,
                          embedding_size=20, embedding_activation='sigmoid',
                          rnn_type='lstm', mix_key='mix_magnitude'):
    """
    This is a builder for an open-unmix LIKE (UMX) architecture for music source
    separation. 
    
    The architecture is not exactly the same but is very similar for the
    most part. This architecture also has the option of having an embedding space
    attached to it, making it a UMX + Chimera network that you can regularize with
    a deep clustering loss.
    
    Args:
        num_features (int): Number of features in the input spectrogram (usually means
          window length of STFT // 2 + 1.)
        hidden_size (int): Hidden size of the RNN. Will be hidden_size // 2 if bidirectional is True.
        num_layers (int): Number of layers in the RNN.
        bidirectional (int): Whether the RNN is bidirectional.
        dropout (float): Amount of dropout to be used between layers of RNN.
        num_sources (int): Number of sources to create masks for.
        num_audio_channels (int): Number of audio channels in input (e.g. mono or stereo).
          Defaults to 1.
        add_embedding (bool): Whether or not to add an embedding layer to this to make this a 
          Chimera network. If True, then ``embedding_size`` and ``embedding_activation`` will
          be used to define this.
        embedding_size (int): Embedding dimensionality of the deep clustering network. 
        embedding_activation (list of str): Activation of the embedding ('sigmoid', 'softmax', etc.). 
          See ``nussl.ml.networks.modules.Embedding``.
        rnn_type (str, optional): RNN type, either 'lstm' or 'gru'. Defaults to 'lstm'.
        mix_key (str, optional): The key to look for in the input dictionary that contains
          the mixture spectrogram. Defaults to 'mix_magnitude'.
    
    Returns:
        dict: An OpenUnmix-like configuration that can be passed to
          SeparationModel.
    """
    # define the building blocks
    modules = {
        mix_key: {},
        'projection': {
            'class': 'Embedding',
            'args': {
                'num_features': hidden_size,
                'hidden_size': num_features * num_audio_channels,
                'embedding_size': 1,
                'activation': 'none',
                'num_audio_channels': 1,
                'bias': False,
                'reshape': False,
                'dim_to_embed': [2, 3]
            }
        },
        'recurrent_stack': {
            'class': 'RecurrentStack',
            'args': {
                'num_features': hidden_size,
                'hidden_size': (
                    hidden_size // 2 
                    if bidirectional 
                    else hidden_size
                ),
                'num_layers': num_layers,
                'bidirectional': bidirectional,
                'dropout': dropout,
                'rnn_type': rnn_type
            }
        },
        'skip_connection': {
            'class': 'Concatenate',
            'args': {
                'dim': -1
            }
        },
        'input_scalar': {
            'class': 'BatchNorm'
        },
        'output_scalar': {
            'class': 'BatchNorm'
        },
        'bn1': {
            'class': 'BatchNorm',
            'args': {
                'num_features': hidden_size
            }
        },
        'bn2': {
            'class': 'BatchNorm',
            'args': {
                'num_features': hidden_size
            }
        },
        'bn3': {
            'class': 'BatchNorm',
            'args': {
                'num_features': num_features
            }
        },
        'tanh_before_lstm': {
            'class': 'Tanh' 
        },
        'mask': {
            'class': 'ReLU'
        },
        'dense_after_lstm': {
            'class': 'Linear',
            'args': {
                'in_features': 2 * hidden_size,
                'out_features': hidden_size,
                'bias': False
            }
        },
        'inverse_projection': {
            'class': 'Embedding',
            'args': {
                'num_features': num_features,
                'hidden_size': hidden_size,
                'embedding_size': num_sources,
                'activation': 'none',
                'num_audio_channels': num_audio_channels,
                'bias': False,
                'reshape': True,
                'dim_to_embed': -1
            }
        },
        'estimates': {
            'class': 'Mask',
        }
    }

    # define the topology
    connections = [
        ['input_scalar', [mix_key]],
        ['projection', ['input_scalar']],
        ['bn1', ['projection']],
        ['tanh_before_lstm', ['bn1']],
        ['recurrent_stack', ['tanh_before_lstm']],
        ['skip_connection', ['recurrent_stack', 'tanh_before_lstm']],
        ['dense_after_lstm', ['skip_connection']],
        ['bn2', ['dense_after_lstm']],
        ['inverse_projection', ['bn2']],
        ['bn3', ['inverse_projection']],
        ['output_scalar', ['bn3']],
        ['mask', ['output_scalar']],
        ['estimates', ['mask', mix_key]]
    ]
    
    # define the outputs
    output = ['estimates', 'mask']
    
    if add_embedding:
        modules['embedding'] = {
            'class': 'Embedding',
            'args': {
                'num_features': num_features,
                'hidden_size': hidden_size,
                'embedding_size': embedding_size,
                'activation': embedding_activation,
                'num_audio_channels': num_audio_channels,
                'bias': True,
                'reshape': True,
                'dim_to_embed': -1
            }
        }
        connections.append(['embedding', ['bn2']])
        output.append('embedding')

    # put it together
    config = {
        'name': 'OpenUnmix',
        'modules': modules,
        'connections': connections,
        'output': output
    }

    return config


def build_recurrent_end_to_end(num_filters, filter_length, hop_length, window_type, 
                               hidden_size, num_layers, bidirectional, dropout, 
                               num_sources, mask_activation, num_audio_channels=1,
                               mask_complex=False, trainable=False, rnn_type='lstm', 
                               mix_key='mix_audio', normalization_class='BatchNorm',
                               normalization_args=None):
    """
    Builds a config for a BLSTM-based network that operates on the time-series. 
    Uses an STFT within the network and can apply the mixture phase to
    the estimate, or can learn a mask on the phase as well as the magnitude.
    
    Args:
        num_filters (int): Number of learnable filters in the front end network.
        filter_length (int): Length of the filters.
        hop_length (int): Hop length between frames.
        window_type (str): Type of windowing function on apply to each frame.
        hidden_size (int): Hidden size of the RNN.
        num_layers (int): Number of layers in the RNN.
        bidirectional (int): Whether the RNN is bidirectional.
        dropout (float): Amount of dropout to be used between layers of RNN.
        num_sources (int): Number of sources to create masks for. 
        mask_activation (list of str): Activation of the mask ('sigmoid', 'softmax', etc.). 
          See ``nussl.ml.networks.modules.Embedding``.
        num_audio_channels (int): Number of audio channels in input (e.g. mono or stereo).
          Defaults to 1.
        mask_complex (bool, optional): Whether to also place a mask on the complex part, or
          whether to just use the mixture phase.
        trainable (bool, optional): Whether to learn the filters, which start from a 
          Fourier basis.
        rnn_type (str, optional): RNN type, either 'lstm' or 'gru'. Defaults to 'lstm'.
        normalization_class (str, optional): Type of normalization to apply, either
          'InstanceNorm' or 'BatchNorm'. Defaults to 'BatchNorm'.
        normalization_args (dict, optional): Args to normalization class, optional.
        mix_key (str, optional): The key to look for in the input dictionary that contains
          the mixture spectrogram. Defaults to 'mix_magnitude'.
    
    Returns:
        dict: A recurrent end-to-end network configuration that can be passed to
          SeparationModel.
    """
    normalization_args = {} if normalization_args is None else normalization_args
    cutoff = num_filters // 2 + 1
    num_features = 2 * cutoff if mask_complex else cutoff

    # define the building blocks
    modules = {
        mix_key: {},
        'audio': {
            'class': 'STFT',
            'args': {
                'num_filters': num_filters,
                'filter_length': filter_length,
                'hop_length': hop_length,
                'window_type': window_type,
                'requires_grad': trainable
            }
        },
        'log_spectrogram': {
            'class': 'AmplitudeToDB'
        },
        'normalization': {
            'class': normalization_class,
            'args': normalization_args,
        },
        'split': {
            'class': 'Split',
            'args': {
                'split_sizes': (cutoff, cutoff),
                'dim': 2
            }
        },
        'concatenate': {
            'class': 'Concatenate',
            'args': {
                'dim': 2
            }
        },
        'expand': {
            'class': 'Expand',
        },
        'recurrent_stack': {
            'class': 'RecurrentStack',
            'args': {
                'num_features': num_features,
                'hidden_size': hidden_size,
                'num_layers': num_layers,
                'bidirectional': bidirectional,
                'dropout': dropout,
                'rnn_type': rnn_type
            }
        },
        'mask': {
            'class': 'Embedding',
            'args': {
                'num_features': num_features,
                'hidden_size': hidden_size * 2 if bidirectional else hidden_size,
                'embedding_size': num_sources,
                'activation': mask_activation,
                'num_audio_channels': num_audio_channels
            }
        },
        'estimates': {
            'class': 'Mask',
        },
    }
    
    # define the topology
    if not mask_complex:
        connections = [
            ['audio', [mix_key, {'direction': 'transform'}]],
            ['split', ['audio',]],
            ['log_spectrogram', ['split:0', ]],
            ['normalization', ['log_spectrogram']],
            ['recurrent_stack', ['normalization', ]],
            ['mask', ['recurrent_stack', ]],
            ['estimates', ['mask', 'split:0']],
            ['expand', ['mask', 'split:1']],
            ['concatenate', ['estimates', 'expand']],
            ['audio', ['concatenate', {'direction': 'inverse'}]]
        ]
    else:
        connections = [
            ['audio', [mix_key, {'direction': 'transform'}]],
            ['normalization', ['audio']],
            ['recurrent_stack', ['normalization', ]],
            ['mask', ['recurrent_stack', ]],
            ['estimates', ['mask', 'audio']],
            ['audio', ['estimates', {'direction': 'inverse'}]]
        ]

    # define the outputs
    output = ['audio', 'mask']

    # put it together
    config = {
        'name': 'EndToEndRNN',
        'modules': modules,
        'connections': connections,
        'output': output
    }

    return config


def build_dual_path_recurrent_end_to_end(
        num_filters, filter_length, hop_length, 
        chunk_size, hop_size, hidden_size, num_layers, 
        bidirectional, bottleneck_size,
        num_sources, mask_activation, num_audio_channels=1,
        window_type='rectangular', skip_connection=False,
        embedding_bias=True, rnn_type='lstm', mix_key='mix_audio',
        init_forget=False):
    """
    Builds a config for a dual path recurrent network that operates on the 
    time-series. Uses a learned filterbank within the network.
    
    Args:
        num_filters (int): Number of learnable filters in the front end network.
        filter_length (int): Length of the filters.
        hop_length (int): Hop length between frames.
        window_type (str): Type of windowing function on apply to each frame.
        hidden_size (int): Hidden size of the RNN.
        num_layers (int): Number of layers in the RNN.
        bidirectional (int): Whether the RNN is bidirectional.
        dropout (float): Amount of dropout to be used between layers of RNN.
        num_sources (int): Number of sources to create masks for. 
        mask_activation (list of str): Activation of the mask ('sigmoid', 'softmax', etc.). 
          See ``nussl.ml.networks.modules.Embedding``.
        num_audio_channels (int): Number of audio channels in input (e.g. mono or stereo).
          Defaults to 1.
        rnn_type (str, optional): RNN type, either 'lstm' or 'gru'. Defaults to 'lstm'.
        normalization_class (str, optional): Type of normalization to apply, either
          'InstanceNorm' or 'BatchNorm'. Defaults to 'BatchNorm'.
        mix_key (str, optional): The key to look for in the input dictionary that contains
          the mixture spectrogram. Defaults to 'mix_magnitude'.
    
    Returns:
        dict: A TASNet configuration that can be passed to
          SeparationModel.
    """
    
    # define the building blocks
    recurrent_hidden_size = hidden_size // 2 if bidirectional else hidden_size
    recurrent_stack = {
        'class': 'RecurrentStack',
        'args': {
            'num_features': bottleneck_size,
            'hidden_size': recurrent_hidden_size,
            'num_layers': 1,
            'bidirectional': bidirectional,
            'dropout': 0.0,
            'rnn_type': rnn_type,
            'batch_first': True,
            'init_forget': init_forget,
        }
    }

    modules = {
        mix_key: {},
        'audio': {
            'class': 'LearnedFilterBank',
            'args': {
                'num_filters': num_filters,
                'filter_length': filter_length,
                'hop_length': hop_length,
                'window_type': window_type,
                'requires_grad': True
            }
        },
        'mixture_weights': {
            'class': 'ReLU'
        },
        'dual_path': {
            'class': 'DualPath',
            'args': {
                'num_layers': num_layers,
                'chunk_size': chunk_size,
                'hop_size': hop_size,
                'skip_connection': skip_connection,
                'in_features': num_filters,
                'bottleneck_size': bottleneck_size,
                # rest are args to DualPathBlock
                'hidden_size': hidden_size,
                'intra_processor': recurrent_stack,
                'inter_processor': recurrent_stack,
            } 
        },
        'mask': {
            'class': 'Embedding',
            'args': {
                'num_features': num_filters,
                'hidden_size': num_filters,
                'embedding_size': num_sources,
                'activation': mask_activation,
                'num_audio_channels': num_audio_channels,
                'bias': embedding_bias,
                'dim_to_embed': [2, 3],
            }
        },
        'estimates': {
            'class': 'Mask',
        },
    }
    
    # define the topology
    connections = [
        ['audio', [mix_key, {'direction': 'transform'}]],
        ['mixture_weights', ['audio']],
        ['dual_path', ['mixture_weights', ]],
        ['mask', ['dual_path', ]],
        ['estimates', ['mask', 'mixture_weights']],
        ['audio', ['estimates', {'direction': 'inverse'}]]
    ]

    # define the outputs
    output = ['audio', 'mask']

    # put it together
    config = {
        'name': 'DualPathRNN',
        'modules': modules,
        'connections': connections,
        'output': output
    }

    return config



import numpy as np

from ... import ml
from ... import AudioSignal


class NMFMixin:
    @staticmethod
    def fit(audio_signals, n_components, beta_loss='frobenius',
            l1_ratio=0.5, **kwargs):
        """
        Fits an NMF model to the magnitude spectrograms of each
        audio signal. If `audio_signals` is a list, the magnitude
        spectrograms of each signal are concatenated into a single
        data matrix to which NMF is fit. If `audio_signals`
        is a single audio signal, then NMF is fit only to the
        magnitude spectrogram for that audio signal. If any of
        the audio signals are multichannel, the channels are 
        concatenated into a single (longer) data matrix.

        Args:
            audio_signals (list or AudioSignal): AudioSignal object(s) that 
              NMF will be fit to.
            n_components (int): Number of components to use in the NMF
              module. Corresponds to number of spectral templates.
            beta_loss (float or string): String must be in 
              {'frobenius', 'kullback-leibler', 'itakura-saito'}.
              Beta divergence to be minimized, measuring the distance between X
              and the dot product WH. Note that values different from 'frobenius'
              (or 2) and 'kullback-leibler' (or 1) lead to significantly slower
              fits. Note that for beta_loss <= 0 (or 'itakura-saito'), the input
              matrix X cannot contain zeros. Used only in 'mu' solver. Defaults to 
              'frobenius'.
            l1_ratio (float): The regularization mixing parameter, with 0 <= l1_ratio <= 1.
              For l1_ratio = 0 the penalty is an elementwise L2 penalty (aka Frobenius Norm).
              For l1_ratio = 1 it is an elementwise L1 penalty.
              For 0 < l1_ratio < 1, the penalty is a combination of L1 and L2.
              Defaults to 1.0 (sparse templates and activations).
            kwargs (dict): Additional keyword arguments to initialization of the NMF
              decomposition method.

        Returns:
            model (NMF): Fitted NMF model to the audio signal(s).
            components (np.ndarray): Spectral templates (n_components, n_features)
            activations (np.ndarray): Activations (n_components, n_time, n_channels)
              The shape here is as if it was like an STFT but with components as the
              features rather than frequencies of the STFT.
        """
        if isinstance(audio_signals, AudioSignal):
            audio_signals = [audio_signals]

        data = []

        n_spectrograms = 0

        for audio_signal in audio_signals:
            _data = np.abs(audio_signal.stft())
            n_spectrograms += audio_signal.num_channels
            # flip around array so frequencies are last
            _data = _data.transpose()
            # flatten first 2 axes
            _data = _data.reshape(-1, _data.shape[-1])
            data.append(_data)

        data = np.concatenate(data, axis=0)

        model = ml.NMF(n_components=n_components, l1_ratio=l1_ratio,
                       beta_loss=beta_loss, **kwargs)
        activations = model.fit_transform(data)
        activations = activations.T.reshape(n_components, -1, n_spectrograms)
        return model, model.components_, activations

    @staticmethod
    def transform(audio_signal, model):
        """
        Use an already fit model to transform the magnitude spectrogram of an 
        audio signal into components and activations. These can be multiplied to 
        reconstruct the original matrix, or used to separate out sounds that correspond
        to components in the model.
        
        Args:
            audio_signal (AudioSignal): AudioSignal object to transform with model.
            model (NMF): NMF model to separate with. Must be fitted prior to this call.
            

        Returns:
            components (np.ndarray): Spectral templates (n_components, n_features)
            activations (np.ndarray): Activations (n_components, n_time, n_channels)
              The shape here is as if it was like an STFT but with components as the
              features rather than frequencies of the STFT.
        """
        data = np.abs(audio_signal.stft())

        shape = data.shape
        data = data.transpose()
        data = data.reshape(-1, data.shape[-1])
        activations = model.transform(data).T

        activations = activations.reshape((model.n_components,) + shape[1:])
        return model.components_, activations

    @staticmethod
    def inverse_transform(components, activations):
        """
        Reconstructs the magnitude spectrogram by matrix multiplying the components 
        with the activations. Components and activations are considered to be 2D matrices, 
        but if they are more, then the first dimension is interpreted to be the batch 
        dimension.
        
        Args:
            components (np.ndarray): Spectral templates (n_components, n_features)
            activations (np.ndarray): Activations (n_components, n_time, n_channels)
              The shape here is as if it was like an STFT but with components as the
              features rather than frequencies of the STFT.
        """
        activations = activations.transpose()
        shape = activations.shape

        reconstruction = activations @ components
        reconstruction = reconstruction.reshape(shape[:-1] + (-1,))

        return reconstruction.transpose()



from ..base import SeparationBase

class MixAsEstimate(SeparationBase):
    """
    This algorithm does nothing but scale the mix by the number of sources. This can
    be used to compute the improvement metrics (e.g. improvement in SDR over using the
    mixture as the estimate).
    
    Args:
        input_audio_signal (AudioSignal): Signal to separate.
        num_sources (int): How many sources to return.
    """

    def __init__(self, input_audio_signal, num_sources):
        self.num_sources = num_sources
        super().__init__(input_audio_signal=input_audio_signal)

    def run(self):
        pass

    def make_audio_signals(self):
        scalar = 1
        estimates = [
            self.audio_signal * scalar for i in range(self.num_sources)
        ]
        return estimates



import nussl
from torch import nn
import torch
from .... import AudioSignal

class FilterBank(nn.Module):
    """
    Base class for implementing short-time filter-bank style transformations
    of an audio signal. 
    
    This class accepts two different tensors, as there are two modes it can
    be called in:

    - transform: This takes an audio signal and maps it to a spectral 
      representation by applying the internal filterbank.

    - inverse: This takes a spectral representation and maps it back to
      the audio domain.

    There are two unimplemented functions in this class that must be
    implemented by the subclass:

    - ``get_transform_filters``: This should produce a filter bank that can be
      applied to the audio signal, of shape ``(filter_length, num_filters)``. The
      filter bank is applied by chunking the signal into overlapping segments
      using ``nn.Unfold``, and performing a matrix multiplication of the chunks 
      (after permuting the dimensions appropriately) with the filter bank.

    - ``get_inverse_filters``: This should produce a filter bank that maps the
      spectral representation back to the audio domain. This is done by first
      permuting the dimensions of the spectral representations appropriately,
      then matrix multiplying the spectral representation with the filter bank.
      Finally, the signal is resynthesized via overlap-add using ``nn.Fold``.

    Windowing is applied to the signal, according to ``window_type``, which
    can be any of the windows found in ``nussl.core.constants.ALL_WINDOWS``.

    This can also be applied to multiple sources at once, if they are on the
    last axis after all of the feature dimensions. If the number of features
    in the data is greater than the expected number by one, then the last
    dimension is assumed to be the source dimension. This dimension is then
    mapped and merged with the batch dimension so the transforms can be applied
    to all of the sources. Before returning the data, the source dimension is 
    moved back to the last dimension.

    In the forward pass, keyword arguments can also be passed through. These
    keyword arguments get passed through to ``apply_filter`` and ``get_filters``,
    in case these should be conditioned on something during the forward pass.

    Note:
        The output dimensionality may not always match what is given by 
        something like scipy.stft (e.g. might be off by one in frames) for 
        some hop lengths! However, an attempt was made and for hop lengths
        that are half the filter length or a quarter of the filter length, the
        number of segments will match that of scipy.stft.
    
    Args:
        num_filters (int): Number of filters in both filterbanks.
        filter_length (int, optional): Length of each filter. Defaults to None,
            equal to num_filters.
        hop_length (int, optional): Hop length between each filter. 
            Defaults to None, half of filter_length.
        window_type (str, optional): Type of window to use. Defaults to 
            'sqrt_hann'.
        dilation (int, optional): Dilation of nn.Unfold and nn.Fold. Could be
            useful for implementing dilated convolutional frontends. 
            Defaults to 1.
        direction (str, optional): In which direction to take the input data. 
            Either 'transform' or 'inverse'. Can also be set during the 
            forward pass. Defaults to 'transform'.
        requires_grad (bool, optional): Whether to make the filterbank learnable
            during backprop. Defaults to False.
    """
    def __init__(self, num_filters, filter_length=None, hop_length=None,
                 window_type='sqrt_hann', dilation=1, direction='transform',
                 requires_grad=False):
        super().__init__()
        
        self.num_filters = num_filters
        self.filter_length = (
            num_filters 
            if filter_length is None
            else filter_length
        )
        self.hop_length = (
            self.filter_length // 2 
            if hop_length is None 
            else hop_length
        )
        
        self.direction = direction
        self.requires_grad = requires_grad
        self.dilation = dilation
        self.window_type = window_type
        self.output_length = None

        self.register_buffer('window', self._get_window())

        self.transform_filters = self.get_transform_filters()
        self.inverse_filters = self.get_inverse_filters()
        
    def _get_window(self):
        window = AudioSignal.get_window(
            self.window_type, self.filter_length
        )
        window = torch.from_numpy(window).float()
        return window.reshape(1, 1, -1, 1)
        
    def get_transform_filters(self):
        raise NotImplementedError()
        
    def get_inverse_filters(self):
        raise NotImplementedError()
    
    def get_filters(self):
        filters = (
            self.transform_filters
            if self.direction == 'transform'
            else self.inverse_filters
        )
        return filters
    
    def apply_filter(self, data, **kwargs):
        filters = self.get_filters(**kwargs)        
        data = data.transpose(-1, 2)
        data = data @ filters
        data = data.transpose(-1, 2)
        return data   
    
    def transform(self, data, **kwargs):
        ndim = data.ndim
        if ndim > 3:
            # move sources to the batch dimension
            # then fix it later
            num_sources = data.shape[-1]
            data = data.permute(0, -1, 1, 2)
            data = data.reshape(-1, *data.shape[2:])

        self.original_length = data.shape[-1]
        pad_extra = (
            (-(data.shape[-1] - self.filter_length) % self.hop_length)
            % self.filter_length
        )
        pad_tuple = (
            self.filter_length // 2,
            self.filter_length // 2 + pad_extra
        )
        data = nn.functional.pad(data, pad_tuple)
        self.output_length = data.shape[-1]
        
        num_batch, num_audio_channels, num_samples = data.shape

        unfold = nn.Unfold(
            kernel_size=(1, self.filter_length), 
            stride=(1, self.hop_length),
            dilation=self.dilation,
            padding=(0, 0)
        )
        data = data.reshape(
            num_batch * num_audio_channels,
            1, 1, num_samples
        )
        data = unfold(data)
        data = data.view(
            num_batch, num_audio_channels,
            self.filter_length, -1
        )
        data = data.permute(0, 3, 2, 1)
        data = data * self.window
        data = self.apply_filter(data, **kwargs)

        if ndim > 3:
            # then we moved sources to the batch dimension
            # we need to move it back before returning
            data = data.reshape(
                -1, num_sources, *data.shape[1:])
            data = data.permute(0, 2, 3, 4, 1)
        return data
    
    def inverse(self, data, **kwargs):
        ndim = data.ndim
        if ndim > 4:
            # move sources to the batch dimension
            # then fix it later
            num_sources = data.shape[-1]
            data = data.permute(0, -1, 1, 2, 3)
            data = data.reshape(-1, *data.shape[2:])
        
        data = self.apply_filter(data, **kwargs)

        data *= self.window.sum()
        data = data * self.window
        
        num_batch, sequence_length, num_features, num_audio_channels = (
            data.shape
        )
        
        data = data.permute(0, 3, 2, 1)
        data = data.reshape(-1, data.shape[2], data.shape[3])

        fold = nn.Fold(
            (1, self.output_length),
            kernel_size=(1, self.filter_length),
            stride=(1, self.hop_length),
            dilation=self.dilation,
            padding=(0, 0),
        )
        norm = data.new_ones(data.shape)
        norm *= self.window.view(1, -1, 1) ** 2
        
        data = fold(data)
        norm = fold(norm)
        norm[norm < 1e-10] = 1
        data = data / norm

        data = data.reshape(num_batch, num_audio_channels, -1)
        
        boundary = self.filter_length // 2
        data = data[..., boundary:-boundary]
        data = data[..., :self.original_length]

        if ndim > 4:
            # then we moved sources to the batch dimension
            # we need to move it back before returning
            data = data.reshape(
                -1, num_sources, num_audio_channels, data.shape[-1])
            data = data.permute(0, 2, 3, 1)

        return data
        
    def forward(self, data, direction=None, **kwargs):
        if direction is not None:
            self.direction = direction
        if self.direction == 'transform':
            func = self.transform
        elif self.direction == 'inverse':
            func = self.inverse
        return func(data, **kwargs)

class STFT(FilterBank):
    """
    An implementation of STFT and iSTFT using nn.Unfold, nn.Fold, and matrix
    multiplication with a Fourier basis.

    The usual way to compute an STFT is to split the signal into overlapping
    chunks, multiply each chunk with a window function, and then 
    apply an FFT to each chunk individually. Here, instead of taking the 
    FFT with something like ``torch.fft``, we instead use the matrix
    formulation of FFT. 

    To resynthesize the signal, we use inverse windowing and the pseudoinverse
    of the FFT matrix as our filterbank. We then use overlap/add and divide by
    a normalization factor to reconstruct the signal. The usual error for this
    implementation is on the order of 1e-7.

    In the `transform` direction, this class returns the magnitude and phase
    of the STFT, concatenated along a single dimension. Generally, the first
    half of this is what you would operate on (the magnitudes), while the
    second half you would keep around for reconstructing the signal later on.
    """
    def apply_filter(self, data):
        if self.direction == 'transform':
            scale = torch.sqrt(1.0 / (self.window.sum() ** 2))
            data = super().apply_filter(data)
            data *= scale
            
            eps = 1e-8
            cutoff = 1 + self.filter_length // 2
            real_part = data[..., :cutoff, :]
            imag_part = data[..., cutoff:, :]
            real_part[real_part.abs() <= eps] = eps
            imag_part[imag_part.abs() <= eps] = eps
            
            magnitude = torch.sqrt(
                real_part ** 2 + imag_part ** 2)
            phase = torch.atan2(imag_part, real_part)      
            data = torch.cat([magnitude, phase], dim=2)

        elif self.direction == 'inverse':
            cutoff = 1 + self.filter_length // 2
            magnitude = data[..., :cutoff, :]
            phase = data[..., cutoff:, :]
            data = torch.cat(
                [
                    magnitude * torch.cos(phase), 
                    magnitude * torch.sin(phase)
                ],
                dim=2
            )
            data = super().apply_filter(data)
            
        return data
    
    def _get_fft_basis(self):
        fourier_basis = torch.rfft(
            torch.eye(self.filter_length), 
            1, onesided=True
        )
        cutoff = 1 + self.filter_length // 2
        fourier_basis = torch.cat([
            fourier_basis[:, :cutoff, 0],
            fourier_basis[:, :cutoff, 1]
        ], dim=1)
        return fourier_basis.float()
    
    def get_transform_filters(self):
        fourier_basis = self._get_fft_basis()
        return nn.Parameter(fourier_basis, requires_grad=self.requires_grad)
    
    def get_inverse_filters(self):
        fourier_basis = self._get_fft_basis()
        inverse_filters = torch.pinverse(
            fourier_basis.unsqueeze(0)).squeeze(0)
        return nn.Parameter(inverse_filters, requires_grad=self.requires_grad)

class LearnedFilterBank(FilterBank):
    """
    Implements a simple learnable filter bank. The filter bank is completely
    random on initialization and should be learned during backprop. 
    `requires_grad` is set to True regardless of what the parent class
    init has.
    """
    def get_transform_filters(self):
        basis = nn.Parameter(
            torch.ones(self.filter_length, self.num_filters),
            requires_grad=True
        )
        nn.init.xavier_normal_(basis)
        return basis

    def get_inverse_filters(self):
        basis = nn.Parameter(
            torch.ones(self.num_filters, self.filter_length),
            requires_grad=True
        )
        nn.init.xavier_normal_(basis)
        return basis



"""
There are ways to measure the quality of a separated source without
requiring ground truth. These functions operate on the output of
clustering-based separation algorithms and work by analyzing
the clusterability of the feature space used to generate the
separated sources.
"""

from sklearn.metrics import silhouette_samples
import numpy as np
from .cluster import KMeans, GaussianMixture
from scipy.special import logsumexp
from .train import loss
import torch

def softmax(x, axis=None):
    return np.exp(x - logsumexp(x, axis=axis, keepdims=True))

def jensen_shannon_divergence(gmm_p, gmm_q, n_samples=10**5):
    """
    Compute Jensen-Shannon (JS) divergence between two Gaussian Mixture Models via 
    sampling. JS divergence is also known as symmetric Kullback-Leibler divergence.
    JS divergence has no closed form in general for GMMs, thus we use sampling to 
    compute it.

    Args:
        gmm_p (GaussianMixture): A GaussianMixture class fit to some data.
        gmm_q (GaussianMixture): Another GaussianMixture class fit to some data.
        n_samples (int): Number of samples to use to estimate JS divergence. 

    Returns:
        JS divergence between gmm_p and gmm_q
    """
    X = gmm_p.sample(n_samples)[0]
    log_p_X = gmm_p.score_samples(X)
    log_q_X = gmm_q.score_samples(X)
    log_mix_X = np.logaddexp(log_p_X, log_q_X)

    Y = gmm_q.sample(n_samples)[0]
    log_p_Y = gmm_p.score_samples(Y)
    log_q_Y = gmm_q.score_samples(Y)
    log_mix_Y = np.logaddexp(log_p_Y, log_q_Y)

    return (log_p_X.mean() - (log_mix_X.mean() - np.log(2))
            + log_q_Y.mean() - (log_mix_Y.mean() - np.log(2))) / 2

def _get_loud_bins_mask(threshold, audio_signal=None, representation=None):
    if representation is None:
        representation = np.abs(audio_signal.stft())
    threshold = np.percentile(representation, threshold)
    mask = representation > threshold
    return mask, representation

def jensen_shannon_confidence(audio_signal, features, num_sources, threshold=95, 
                              n_samples=10**5, **kwargs):
    """
    Calculates the clusterability of a space by comparing a K-cluster GMM
    with a 1-cluster GMM on the same features. This function fits two
    GMMs to all of the points that are above the specified threshold (defaults
    to 95: 95th percentile of all the data). This saves on computation time and
    also allows one to have the confidence measure only focus on the louder
    more perceptually important points.

    References:

    Seetharaman, Prem, Gordon Wichern, Jonathan Le Roux, and Bryan Pardo. 
    “Bootstrapping Single-Channel Source Separation via Unsupervised Spatial 
    Clustering on Stereo Mixtures”. 44th International Conference on Acoustics, 
    Speech, and Signal Processing, Brighton, UK, May, 2019

    Seetharaman, Prem. Bootstrapping the Learning Process for Computer Audition. 
    Diss. Northwestern University, 2019.
    
    Args:
        audio_signal (AudioSignal): AudioSignal object which will be used to compute
          the mask over which to compute the confidence measure. This can be None, if
          and only if ``representation`` is passed as a keyword argument to this 
          function.
        features (np.ndarray): Numpy array containing the features to be clustered. 
          Should have the same dimensions as the representation.
        n_sources (int): Number of sources to cluster the features into.
        threshold (int, optional): Threshold by loudness. Points below the threshold are
          excluded from being used in the confidence measure. Defaults to 95.
        kwargs: Keyword arguments to `_get_loud_bins_mask`. Namely, representation can
          go here as a keyword argument.

    Returns:
        float: Confidence given by Jensen-Shannon divergence.
    """
    mask, _ = _get_loud_bins_mask(threshold, audio_signal, **kwargs)
    embedding_size = features.shape[-1]
    features = features[mask].reshape(-1, embedding_size)

    one_component_gmm = GaussianMixture(1)
    n_component_gmm = GaussianMixture(num_sources)

    one_component_gmm.fit(features)
    n_component_gmm.fit(features)

    confidence = jensen_shannon_divergence(
        one_component_gmm, n_component_gmm, n_samples=n_samples)

    return confidence

def posterior_confidence(audio_signal, features, num_sources, threshold=95, 
                         **kwargs):
    """
    Calculates the clusterability of an embedding space by looking at the
    strength of the assignments of each point to a specific cluster. The 
    more points that are "in between" clusters (e.g. no strong assignmment),
    the lower the clusterability.

    References:

    Seetharaman, Prem, Gordon Wichern, Jonathan Le Roux, and Bryan Pardo. 
    “Bootstrapping Single-Channel Source Separation via Unsupervised Spatial 
    Clustering on Stereo Mixtures”. 44th International Conference on Acoustics, 
    Speech, and Signal Processing, Brighton, UK, May, 2019

    Seetharaman, Prem. Bootstrapping the Learning Process for Computer Audition. 
    Diss. Northwestern University, 2019.
    
    Args:
        audio_signal (AudioSignal): AudioSignal object which will be used to compute
          the mask over which to compute the confidence measure. This can be None, if
          and only if ``representation`` is passed as a keyword argument to this 
          function.
        features (np.ndarray): Numpy array containing the features to be clustered. 
          Should have the same dimensions as the representation.
        n_sources (int): Number of sources to cluster the features into.
        threshold (int, optional): Threshold by loudness. Points below the threshold are
          excluded from being used in the confidence measure. Defaults to 95.
        kwargs: Keyword arguments to `_get_loud_bins_mask`. Namely, representation can
          go here as a keyword argument.
    
    Returns:
        float: Confidence given by posteriors.
    """
    mask, _ = _get_loud_bins_mask(threshold, audio_signal, **kwargs)
    embedding_size = features.shape[-1]
    features = features[mask].reshape(-1, embedding_size)

    kmeans = KMeans(num_sources)
    distances = kmeans.fit_transform(features)

    confidence = softmax(-distances, axis=-1)

    confidence = (
        (num_sources * np.max(confidence, axis=-1) - 1) / 
        (num_sources - 1)
    )

    return confidence.mean()

def silhouette_confidence(audio_signal, features, num_sources, threshold=95, 
                          max_points=1000, **kwargs):
    """
    Uses the silhouette score to compute the clusterability of the feature space.

    The Silhouette Coefficient is calculated using the 
    mean intra-cluster distance (a) and the mean nearest-cluster distance (b) 
    for each sample. The Silhouette Coefficient for a sample is (b - a) / max(a, b). 
    To clarify, b is the distance between a sample and the nearest cluster 
    that the sample is not a part of. Note that Silhouette Coefficient is 
    only defined if number of labels is 2 <= n_labels <= n_samples - 1.

    References:

    Seetharaman, Prem. Bootstrapping the Learning Process for Computer Audition. 
    Diss. Northwestern University, 2019.

    Peter J. Rousseeuw (1987). “Silhouettes: a Graphical Aid to the 
    Interpretation and Validation of Cluster Analysis”. Computational and 
    Applied Mathematics 20: 53-65.
    
    Args:
        audio_signal (AudioSignal): AudioSignal object which will be used to compute
          the mask over which to compute the confidence measure. This can be None, if
          and only if ``representation`` is passed as a keyword argument to this 
          function.
        features (np.ndarray): Numpy array containing the features to be clustered. 
          Should have the same dimensions as the representation.
        n_sources (int): Number of sources to cluster the features into.
        threshold (int, optional): Threshold by loudness. Points below the threshold are
          excluded from being used in the confidence measure. Defaults to 95.
        kwargs: Keyword arguments to `_get_loud_bins_mask`. Namely, representation can
          go here as a keyword argument.
        max_points (int, optional): Maximum number of points to compute the Silhouette
          score for. Silhouette score is a costly operation. Defaults to 1000.
    
    Returns:
        float: Confidence given by Silhouette score.
    """
    mask, _ = _get_loud_bins_mask(threshold, audio_signal, **kwargs)
    embedding_size = features.shape[-1]
    features = features[mask].reshape(-1, embedding_size)

    if features.shape[0] > max_points:
        idx = np.random.choice(
            np.arange(features.shape[0]), max_points,
            replace=False)
        features = features[idx]
    
    kmeans = KMeans(num_sources)

    labels = kmeans.fit_predict(features)
    confidence = silhouette_samples(features, labels)

    return confidence.mean()

def loudness_confidence(audio_signal, features, num_sources, threshold=95, 
                        **kwargs):
    """
    Computes the clusterability of the feature space by comparing the absolute
    size of each cluster.
    
    References:

    Seetharaman, Prem, Gordon Wichern, Jonathan Le Roux, and Bryan Pardo. 
    “Bootstrapping Single-Channel Source Separation via Unsupervised Spatial 
    Clustering on Stereo Mixtures”. 44th International Conference on Acoustics, 
    Speech, and Signal Processing, Brighton, UK, May, 2019

    Seetharaman, Prem. Bootstrapping the Learning Process for Computer Audition. 
    Diss. Northwestern University, 2019.
    
    Args:
        audio_signal (AudioSignal): AudioSignal object which will be used to compute
          the mask over which to compute the confidence measure. This can be None, if
          and only if ``representation`` is passed as a keyword argument to this 
          function.
        features (np.ndarray): Numpy array containing the features to be clustered. 
          Should have the same dimensions as the representation.
        n_sources (int): Number of sources to cluster the features into.
        threshold (int, optional): Threshold by loudness. Points below the threshold are
          excluded from being used in the confidence measure. Defaults to 95.
        kwargs: Keyword arguments to `_get_loud_bins_mask`. Namely, representation can
          go here as a keyword argument.
    
    Returns:
        float: Confidence given by size of smallest cluster.
    """
    mask, _ = _get_loud_bins_mask(threshold, audio_signal, **kwargs)
    embedding_size = features.shape[-1]
    features = features[mask].reshape(-1, embedding_size)

    kmeans = KMeans(num_sources)
    labels = kmeans.fit_predict(features)

    source_shares = np.array(
        [(labels == i).sum() for i in range(num_sources)]
    ).astype(float)
    source_shares *= (1 / source_shares.sum())
    confidence = source_shares.min()

    return confidence

def whitened_kmeans_confidence(audio_signal, features, num_sources, threshold=95, 
                               **kwargs):
    """
    Computes the clusterability in two steps:

    1. Cluster the feature space using KMeans into assignments
    2. Compute the Whitened K-Means loss between the features and the assignments.
    
    Args:
        audio_signal (AudioSignal): AudioSignal object which will be used to compute
          the mask over which to compute the confidence measure. This can be None, if
          and only if ``representation`` is passed as a keyword argument to this 
          function.
        features (np.ndarray): Numpy array containing the features to be clustered. 
          Should have the same dimensions as the representation.
        n_sources (int): Number of sources to cluster the features into.
        threshold (int, optional): Threshold by loudness. Points below the threshold are
          excluded from being used in the confidence measure. Defaults to 95.
        kwargs: Keyword arguments to `_get_loud_bins_mask`. Namely, representation can
          go here as a keyword argument.
    
    Returns:
        float: Confidence given by whitened k-means loss.
    """
    mask, representation = _get_loud_bins_mask(threshold, audio_signal, **kwargs)
    embedding_size = features.shape[-1]
    features = features[mask].reshape(-1, embedding_size)
    weights = representation[mask].reshape(-1)

    kmeans = KMeans(num_sources)
    distances = kmeans.fit_transform(features)
    assignments = (distances == distances.max(axis=-1, keepdims=True))

    loss_func = loss.WhitenedKMeansLoss()

    features = torch.from_numpy(features).unsqueeze(0).float()
    assignments = torch.from_numpy(assignments).unsqueeze(0).float()
    weights = torch.from_numpy(weights).unsqueeze(0).float()

    loss_val = loss_func(features, assignments, weights).item()
    upper_bound = embedding_size + num_sources
    confidence = 1 - (loss_val / upper_bound)
    return confidence

def dpcl_classic_confidence(audio_signal, features, num_sources, threshold=95, 
                            **kwargs):
    """
    Computes the clusterability in two steps:

    1. Cluster the feature space using KMeans into assignments
    2. Compute the classic deep clustering loss between the features and the assignments.
    
    Args:
        audio_signal (AudioSignal): AudioSignal object which will be used to compute
          the mask over which to compute the confidence measure. This can be None, if
          and only if ``representation`` is passed as a keyword argument to this 
          function.
        features (np.ndarray): Numpy array containing the features to be clustered. 
          Should have the same dimensions as the representation.
        n_sources (int): Number of sources to cluster the features into.
        threshold (int, optional): Threshold by loudness. Points below the threshold are
          excluded from being used in the confidence measure. Defaults to 95.
        kwargs: Keyword arguments to `_get_loud_bins_mask`. Namely, representation can
          go here as a keyword argument.
    
    Returns:
        float: Confidence given by deep clustering loss.
    """
    mask, representation = _get_loud_bins_mask(threshold, audio_signal, **kwargs)
    embedding_size = features.shape[-1]
    features = features[mask].reshape(-1, embedding_size)
    weights = representation[mask].reshape(-1)

    kmeans = KMeans(num_sources)
    distances = kmeans.fit_transform(features)
    assignments = (distances == distances.max(axis=-1, keepdims=True))

    loss_func = loss.DeepClusteringLoss()

    features = torch.from_numpy(features).unsqueeze(0).float()
    assignments = torch.from_numpy(assignments).unsqueeze(0).float()
    weights = torch.from_numpy(weights).unsqueeze(0).float()

    loss_val = loss_func(features, assignments, weights).item()
    confidence = 1 - loss_val
    return confidence



import pandas as pd
import json
import termtables
import numpy as np
import os
import textwrap
import copy


def truncate(values, decs=2):
    return np.trunc(values*10**decs)/(10**decs)


def aggregate_score_files(json_files, aggregator=np.nanmedian):
    """
    Takes a list of json files output by an Evaluation method in nussl
    and aggregates all the metrics into a Pandas dataframe. Sample
    output:

    .. code-block:: none

                                SDR        SIR        SAR
        drums  oracle0.json   9.086025  15.025801  10.362709
               random0.json  -6.539877  -6.087538   3.508338
               oracle1.json   9.591432  14.335700  11.365882
               random1.json  -1.358840  -0.993666   9.577297
        bass   oracle0.json   7.936720  12.843092   9.631929
               random0.json  -4.190299  -3.730649   5.802003
               oracle1.json   8.581090  12.513445  10.831370
               random1.json   0.365171   0.697621  11.693103
        other  oracle0.json   2.024207   6.133359   4.158805
               random0.json  -9.857085  -9.481909   0.965199
               oracle1.json   3.961383   6.861785   7.085745
               random1.json  -4.042277  -3.707997   7.260934
        vocals oracle0.json  12.169686  16.650161  14.085037
               random0.json  -2.440166  -1.884026   6.760966
               oracle1.json  12.409913  16.248470  14.725983
               random1.json   1.609577   1.958037  12.738970
    
    Args:
        json_files (list): List of JSON files that will be parsed for metrics.
        aggregator ([type], optional): How to aggregate results within a single
          track. Defaults to np.median.
    
    Returns:
        pd.DataFrame: Pandas dataframe containing the aggregated metrics.
    """
    metrics = {}
    for json_file in json_files:
        with open(json_file, 'r') as f:
            data = json.load(f)
        json_key = os.path.basename(json_file)
        for name in data:
            if name not in ['combination', 'permutation', 'metadata']:
                if name not in metrics:
                    metrics[name] = {}
                if json_key not in metrics[name]:
                    metrics[name][json_key] = {}
                for key in data[name]:
                    _data = aggregator(data[name][key])
                    metrics[name][json_key][key] = _data
    
    df = pd.concat({
        k: pd.DataFrame(v).T for k, v in metrics.items()
    }, axis=0, names=['source', 'file'])
    df.reset_index(inplace=True) 
    
    return df


def _get_mean_and_std(df, decs=2):
    """
    Gets the mean and standard deviation of each metric in the pandas
    DataFrame and returns it as a list of strings.
    """
    excluded_columns = ['source', 'file']

    metrics = [x for x in list(df.columns) if x not in excluded_columns]
    metrics.insert(0, '#')

    means = [
        f'{truncate(m, decs=decs):{4+decs}.{decs}f}' 
        for m in np.array(df.mean()).T
    ]
    stds = [
        f'{truncate(s, decs=decs):{3+decs}.{decs}f}' 
        for s in np.array(df.std()).T
    ]
    data = [f'{m} +/- {s}' for m, s in zip(means, stds)]
    data.insert(0, df.shape[0])

    return metrics, data


def _get_medians(df, decs=2):
    """
    Gets the median of each metric in the pandas
    DataFrame and returns it as a list of strings.
    """
    excluded_columns = ['source', 'file']

    metrics = [x for x in list(df.columns) if x not in excluded_columns]
    metrics.insert(0, '#')

    data = [
        f'{truncate(m, decs=decs):{4+decs}.{decs}f}' 
        for m in np.array(df.median()).T
    ]
    data.insert(0, df.shape[0])
    return metrics, data


def _format_title(title, length, marker=" "):
    pad = (length - len(title)) // 2
    pad = ''.join([marker for _ in range(pad)])
    border = pad + title + pad
    if len(title) % 2:
        border = border + marker
    return border


def _get_report_card(df, func, report_each_source=True, decs=2):
    """
    Gets a report card for a DataFrame using a specific function.
    """
    labels, data = func(df, decs=decs)

    data.insert(0, 'OVERALL')
    data = [data]

    if report_each_source:
        for name in np.unique(df['source']):
            _df = df[df['source'] == name]
            _, _data = func(_df, decs=decs)
            _data.insert(0, name.upper())
            data.append(_data)
    
    # transposing data so each column has the source and its metrics
    data = list(map(list, zip(*data)))
    header = data.pop(0)
    header.insert(0, 'METRIC')
    for i in range(1, len(header)):
        header[i] = _format_title(header[i], 16)
    for l, d in zip(labels, data):
        d.insert(0, l)
        
    alignment = ["c" for _ in header]
    alignment[0] = "l"
    alignment = ''.join(alignment)
    
    report_card = termtables.to_string(
        data, header=header, padding=(0, 1), alignment=alignment)

    return report_card


def report_card(df, notes=None, report_each_source=True, decimals=2):
    """
    Given a Pandas dataframe, usually the output of ``aggregate_score_files``,
    returns a string that looks like this::

    .. code-block:: none
                                                      
                                                                           
                            MEAN +/- STD OF METRICS                       
                                                                            
        ┌─────────┬──────────────────┬──────────────────┬──────────────────┐
        │ METRIC  │     OVERALL      │        S1        │        S2        │
        ╞═════════╪══════════════════╪══════════════════╪══════════════════╡
        │ #       │       6000       │       3000       │       3000       │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SI-SDR  │   11.2 +/-  3.8  │   12.5 +/-  3.5  │    9.8 +/-  3.5  │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SI-SIR  │   22.7 +/-  5.4  │   22.9 +/-  5.0  │   22.6 +/-  5.7  │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SI-SAR  │   11.6 +/-  3.6  │   13.0 +/-  3.3  │   10.1 +/-  3.3  │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SD-SDR  │   10.8 +/-  4.0  │   12.2 +/-  3.8  │    9.3 +/-  3.7  │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SNR     │   11.6 +/-  3.3  │   12.9 +/-  3.1  │   10.3 +/-  3.0  │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SRR     │   22.8 +/-  6.5  │   25.5 +/-  6.3  │   20.0 +/-  5.6  │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SI-SDRi │   11.2 +/-  3.7  │   10.0 +/-  3.4  │   12.3 +/-  3.6  │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SD-SDRi │   10.8 +/-  3.8  │    9.7 +/-  3.6  │   11.8 +/-  3.7  │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SNRi    │   11.6 +/-  3.3  │   10.3 +/-  3.0  │   12.9 +/-  3.1  │
        └─────────┴──────────────────┴──────────────────┴──────────────────┘
                                                                            
                                MEDIAN OF METRICS                          
                                                                            
        ┌─────────┬──────────────────┬──────────────────┬──────────────────┐
        │ METRIC  │     OVERALL      │        S1        │        S2        │
        ╞═════════╪══════════════════╪══════════════════╪══════════════════╡
        │ #       │       6000       │       3000       │       3000       │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SI-SDR  │       11.6       │       13.1       │       10.4       │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SI-SIR  │       23.6       │       23.6       │       23.6       │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SI-SAR  │       12.0       │       13.5       │       10.6       │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SD-SDR  │       11.3       │       12.9       │       10.0       │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SNR     │       11.9       │       13.3       │       10.7       │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SRR     │       23.4       │       26.5       │       20.6       │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SI-SDRi │       11.6       │       10.5       │       12.9       │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SD-SDRi │       11.3       │       10.3       │       12.5       │
        ├─────────┼──────────────────┼──────────────────┼──────────────────┤
        │ SNRi    │       11.9       │       10.7       │       13.3       │
        └─────────┴──────────────────┴──────────────────┴──────────────────┘

                                                                        
                                    NOTES                               
                                                                        
        Uses scale-invariant BSSEval metrics. Evaluated on WSJ0-2Mix at
        8000 Hz sample rate.
    
    Args:
        df (pandas.DataFrame): DataFrame containing the metrics computed during
          evaluation.
        notes (str, optional): Any additional notes you want to be printed at the
          bottom of the report card. Defaults to None.
        report_each_source (bool, optional): Whether or not to report the metrics
          for each individual source type. Defaults to True.
        decimals (int): Number of decimal places to display.
    
    Returns:
        str: A report card for your experiment.
    """
    mean_report_card = _get_report_card(
        df, _get_mean_and_std, report_each_source=report_each_source, decs=decimals)
    median_report_card = _get_report_card(
        df, _get_medians, report_each_source=report_each_source, decs=decimals)

    line_break = mean_report_card.index('\n')

    report_card = (
        f"{_format_title('', line_break)}\n"
        f"{_format_title(' MEAN +/- STD OF METRICS ', line_break)}\n"
        f"{_format_title('', line_break)}\n"
        f"{mean_report_card}\n"
        f"{_format_title('', line_break)}\n"
        f"{_format_title(' MEDIAN OF METRICS ', line_break)}\n"
        f"{_format_title('', line_break)}\n"
        f"{median_report_card}\n"
    )
    
    if notes is not None:
        notes = '\n'.join(textwrap.wrap(notes, line_break))
        report_card += (
            f"{_format_title('', line_break)}\n"
            f"{_format_title(' NOTES ', line_break)}\n"
            f"{_format_title('', line_break)}\n"
            f"{notes}"
        )
    return report_card


def associate_metrics(separation_model, df, test_dataset):
    """
    For a given pandas dataframe (the output of ``aggregate_score_files``), this
    function will associate the high level summary statistics with a model.

    Args:
        separation_model (SeparationModel): A separation object that will have the metrics
            associated with it.
        df (pandas.DataFrame): DataFrame containing the metrics computed during
            evaluation.
        test_dataset (BaseDataset): A dataset object used for the evaluation of the
            metrics.

    Returns:
        (SeparationBase)
    """
    excluded_columns = ['source', 'file']
    metrics = [x for x in list(df.columns) if x not in excluded_columns]
    results = {
        m: {
            'mean': np.mean(df[m]),
            'median': np.median(df[m]),
            'std': np.std(df[m])
        }
        for m in metrics
    }
    separation_model.metadata['evaluation'] = results
    separation_model.metadata['test_dataset'] = copy.deepcopy(test_dataset.metadata)
    return separation_model